// test program for Be condition variables, derived from the book
// _Programming With POSIX Threads_, David R. Butenhof,
// example 3.3.4

#include "condvar.h"
#include <kernel/OS.h>
#include <stdio.h>
#include "errors.h"

struct alarm_t
{
	alarm_t*	link;
	int				seconds;
	bigtime_t		time;
	char				message[64];
};

be_mutex_t	alarm_mutex;
be_cond_t		alarm_cond;
alarm_t*			alarm_list = NULL;
bigtime_t			current_alarm = 0;

// the alarm_insert routine requires that the caller have locked the alarm mutex!
void alarm_insert( alarm_t* alarm )
{
	status_t status;
	alarm_t **last, *next;

	last = &alarm_list;
	next = *last;
	while (next)
	{
		if (next->time >= alarm->time)
		{
			alarm->link = next;
			*last = alarm;
			break;
		}
		last = &next->link;
		next = next->link;
	}

	// if we reached the end of the list, insert the new alarm there
	if (next == NULL)
	{
		*last = alarm;
		alarm->link = NULL;
	}

#ifdef DEBUG
	printf("[list: ");
	for ( next = alarm_list; next != NULL; next = next->link)
	{
		printf("%d(%d)[\"%s\"] ", next->time, next->time - system_time(), next->message);
	}
	printf("]\n");
#endif

	// wake the alarm thread if it is not busy
	if (current_alarm == 0 || alarm->time < current_alarm)
	{
		current_alarm = alarm->time;
		status = be_cond_signal(&alarm_cond);
		if (status != 0)
			err_abort(status, "Signal cond");
	}
}

int32 alarm_thread(void* arg)
{
	alarm_t*		alarm;
	bigtime_t		cond_time;
	bigtime_t		now;
	status_t		status;
	bool				expired;

	// loop forever, processing commands
	be_mutex_lock(&alarm_mutex);
	while (true)
	{
		// if the alarm list is empty, wait until one is added
		current_alarm = 0;
		while (alarm_list == NULL)
		{
			status = be_cond_wait(&alarm_cond, &alarm_mutex);
			if (status)
				err_abort(status, "Wait on cond");
		}
		alarm = alarm_list;
		alarm_list = alarm->link;
		now = system_time();
		expired = false;
		if (alarm->time > now)
		{
#ifdef DEBUG
			printf("[waiting: %d(%d)\"%s\"]\n", alarm->time, alarm->time - system_time(), alarm->message);
#endif
			cond_time = alarm->time;
			current_alarm = alarm->time;
			while (current_alarm == alarm->time)
			{
				status = be_cond_timedwait(&alarm_cond, &alarm_mutex, cond_time);
				if (status == B_TIMED_OUT)
				{
					expired = true;
					break;
				}
				if (status)
					err_abort(status, "Cond timedwait");
			}
			if (!expired)
				alarm_insert(alarm);
		}
		else expired = true;
		if (expired)
		{
			printf("(%d) %s\n", alarm->seconds, alarm->message);
			free(alarm);
		}
	}
}

int main(int, char**)
{
//	status_t		status;
	char				line[128];
	alarm_t*		alarm;
	thread_id		thread;

	be_mutex_init(&alarm_mutex, NULL);
	be_cond_init(&alarm_cond, NULL);

	thread = spawn_thread(alarm_thread, "alarm thread", B_NORMAL_PRIORITY, NULL);
	resume_thread(thread);

	while (true)
	{
		printf("Alarm> ");
		if (fgets(line, sizeof(line), stdin) == NULL) exit(0);
		if (strlen(line) <= 1) continue;
		alarm = (alarm_t*) malloc(sizeof(alarm_t));
		if (!alarm)
			errno_abort("Allocate alarm");

		// parse input line into seconds (%d) and a message (64 non-newline chars)
		if (sscanf(line, "%d %64[^\n]", &alarm->seconds, alarm->message) < 2)
		{
			fprintf(stderr, "Bad command\n");
			free(alarm);
		}
		else
		{
			alarm->time = system_time() + 1000000LL * alarm->seconds;	// convert to usecs
			be_mutex_lock(&alarm_mutex);
			alarm_insert(alarm);			// alarm mutex must be locked here
			be_mutex_unlock(&alarm_mutex);
		}
	}

	return 0;
}
