// BeOS condition variable compatibility layer

#include "condvar.h"
#include <kernel/OS.h>
#include <stdio.h>

// mutex operations
static void lazy_init_mutex(be_mutex_t* mutex)
{
	int32 v = atomic_or(&mutex->init, 1);
	if (2000 == v)			// we're the first, so do the init
	{
		be_mutex_init(mutex, NULL);
	}
	else							// we're not the first, so wait until the init is finished
	{
		while (mutex->init != 9999) snooze(10000);
	}
}

status_t be_mutex_init(be_mutex_t* mutex, be_mutexattr_t* attr)
{
	if (!mutex) return B_BAD_VALUE;
	if (mutex->init == 9999) return EALREADY;

	mutex->lock = create_sem(1, "BeMutex");
	mutex->owner = -1;
	mutex->init = 9999;
	return B_OK;
}

status_t be_mutex_destroy(be_mutex_t* mutex)
{
	if (!mutex) return B_BAD_VALUE;
	if (mutex->init != 9999) return B_NO_INIT;
	delete_sem(mutex->lock);
	return B_OK;
}

status_t be_mutex_lock(be_mutex_t* mutex)
{
	status_t err;

	if (!mutex) return B_BAD_VALUE;
	if (mutex->init < 2000) return B_NO_INIT;
	lazy_init_mutex(mutex);

	err = acquire_sem(mutex->lock);
	if (!err) mutex->owner = find_thread(NULL);
	return err;
}

status_t be_mutex_unlock(be_mutex_t* mutex)
{
	if (!mutex) return B_BAD_VALUE;
	if (mutex->init < 2000) return B_NO_INIT;
	lazy_init_mutex(mutex);

	if (mutex->owner != find_thread(NULL)) return ENOLCK;
	mutex->owner = -1;
	release_sem(mutex->lock);
	return B_OK;
}

//#pragma mark -

// condition variable operations

static void lazy_init_cond(be_cond_t* condvar)
{
	int32 v = atomic_or(&condvar->init, 1);
	if (2000 == v)			// we're the first, so do the init
	{
		be_cond_init(condvar, NULL);
	}
	else							// we're not the first, so wait until the init is finished
	{
		while (condvar->init != 9999) snooze(10000);
	}
}

status_t be_cond_init(be_cond_t* condvar, be_condattr_t* attr)
{
	if (!condvar) return B_BAD_VALUE;
	if (condvar->init == 9999) return EALREADY;

	condvar->sem = create_sem(0, "CVSem");
	condvar->handshakeSem = create_sem(0, "CVHandshake");
	condvar->signalSem = create_sem(1, "CVSignal");
	condvar->ns = condvar->nw = 0;
	condvar->init = 9999;
	return B_OK;
}

status_t be_cond_destroy(be_cond_t* condvar)
{
	if (!condvar) return B_BAD_VALUE;
	if (condvar->init < 2000) return B_NO_INIT;

	condvar->init = 0;
	delete_sem(condvar->sem);
	delete_sem(condvar->handshakeSem);
	delete_sem(condvar->signalSem);
	return B_OK;
}

// wait on a condition variable -- must be called with the mutex held
status_t be_cond_wait(be_cond_t* condvar, be_mutex_t* mutex)
{
	status_t err;

	if (!condvar) return B_BAD_VALUE;
	if (!mutex) return B_BAD_VALUE;
	if (condvar->init < 2000) return B_NO_INIT;
	lazy_init_cond(condvar);

	// serialize access to the counts
	while (acquire_sem(condvar->signalSem) == B_INTERRUPTED) {}
	condvar->nw += 1;
	release_sem(condvar->signalSem);
	be_mutex_unlock(mutex);

	// actually wait for a signal
	err = acquire_sem(condvar->sem);

	// we just awoke, either via a signal or by being interrupted.  If there's
	// a signaller running, he'll think he needs to handshake whether or not
	// we actually awoke due to his signal.  So, we reacquire the signalSem
	// mutex, and handshake if there's a positive signaller count.
	while (acquire_sem(condvar->signalSem) == B_INTERRUPTED) {}
	if (condvar->ns > 0) release_sem(condvar->handshakeSem);
	condvar->nw -= 1;
	condvar->ns -= 1;
	release_sem(condvar->signalSem);

	// always reacquire the mutex before returning, even in error cases
	while (be_mutex_lock(mutex) == B_INTERRUPTED) {}
	return err;
}

// wait for a signal, with a timeout.  as with POSIX condition variables,
// the timeout is in ABSOLUTE time -- it's not a number of microseconds
// to wait, it's the system_time() at which to awaken.
status_t be_cond_timedwait(be_cond_t* condvar, be_mutex_t* mutex, bigtime_t expiration)
{
	status_t err;

	if (!condvar) return B_BAD_VALUE;
	if (!mutex) return B_BAD_VALUE;
	if (condvar->init < 2000) return B_NO_INIT;
	lazy_init_cond(condvar);

	while (acquire_sem(condvar->signalSem) == B_INTERRUPTED) {}
	condvar->nw += 1;
	release_sem(condvar->signalSem);
	be_mutex_unlock(mutex);

	err = acquire_sem_etc(condvar->sem, 1, B_ABSOLUTE_TIMEOUT, expiration);

	while (acquire_sem(condvar->signalSem) == B_INTERRUPTED) {}
	if (condvar->ns > 0) release_sem(condvar->handshakeSem);
	condvar->ns -= 1;
	condvar->nw -= 1;
	release_sem(condvar->signalSem);

	while (be_mutex_lock(mutex) == B_INTERRUPTED) {}
	return err;
}

// signal one waiter
status_t be_cond_signal(be_cond_t* condvar)
{
	if (!condvar) return B_BAD_VALUE;
	if (condvar->init < 2000) return B_NO_INIT;
	lazy_init_cond(condvar);

	while (acquire_sem(condvar->signalSem) == B_INTERRUPTED) {}

	// are there waiters to be awakened?
	if (condvar->nw > condvar->ns)
	{
		condvar->ns += 1;	// inform the next awakening waiter that we need a handshake
		release_sem(condvar->sem);
		release_sem(condvar->signalSem);
		while (acquire_sem(condvar->handshakeSem) == B_INTERRUPTED) {}
	}
	else
	{
		release_sem(condvar->signalSem);
	}
	return B_OK;
}

// signal all waiters
status_t be_cond_broadcast(be_cond_t* condvar)
{
	int32 handshakes;

	if (!condvar) return B_BAD_VALUE;
	if (condvar->init < 2000) return B_NO_INIT;
	lazy_init_cond(condvar);

	while (acquire_sem(condvar->signalSem) == B_INTERRUPTED) {}

	// are there waiters to be awakened?
	if (condvar->nw > condvar->ns)
	{
		handshakes = condvar->nw - condvar->ns;		// condvar->nw will be invalid after releasing signalSem
		condvar->ns = condvar->nw;								// all waiters are accounted for now
		release_sem_etc(condvar->sem, handshakes, 0);
		release_sem(condvar->signalSem);

		// wait for all handshakes
		while (acquire_sem_etc(condvar->handshakeSem, handshakes, 0, 0) == B_INTERRUPTED) {}
	}
	else
	{
		release_sem(condvar->signalSem);
	}
	return B_OK;
}
