#ifndef _PTHREAD_H_
#define _PTHREAD_H_

/*
 * Header files.
 */
/*#include <cdefs.h>*/
/*#include <stdlib.h>*/
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <limits.h>
#include <OS.h>

struct timespec { 
	long tv_sec;
	long tv_usec;
};

/*
 * Run-time invariant values:
 */
#define PTHREAD_DESTRUCTOR_ITERATIONS		4
#define PTHREAD_KEYS_MAX			256
#define PTHREAD_STACK_MIN			1024
#define PTHREAD_THREADS_MAX			ULONG_MAX

/*
 * Compile time symbolic constants for portability specifications:
 *
 * Note that those commented out are not currently supported by the
 * implementation.
 */
#define _POSIX_THREADS
#define _POSIX_THREAD_ATTR_STACKADDR
#define _POSIX_THREAD_ATTR_STACKSIZE
#define _POSIX_THREAD_SAFE_FUNCTIONS

/*
 * Flags for threads and thread attributes.
 */
#define PTHREAD_DETACHED            0x1
#define PTHREAD_SCOPE_SYSTEM        0x2
#define PTHREAD_INHERIT_SCHED       0x4
#define PTHREAD_NOFLOAT             0x8

#define PTHREAD_CREATE_DETACHED     PTHREAD_DETACHED
#define PTHREAD_CREATE_JOINABLE     0
#define PTHREAD_SCOPE_PROCESS       0
#define PTHREAD_EXPLICIT_SCHED      0

/*
 * Flags for read/write lock attributes
 */
#define PTHREAD_PROCESS_PRIVATE     0
#define PTHREAD_PROCESS_SHARED      1	

/*
 * Forward structure definitions.
 *
 * These are mostly opaque to the user.
 */
struct pthread;
struct pthread_attr;
struct pthread_cond;
struct pthread_cond_attr;
struct pthread_mutex;
struct pthread_mutex_attr;
struct pthread_once;
struct pthread_rwlock;
struct pthread_rwlockattr;
struct sched_param;

/*
 * Primitive system data type definitions required by P1003.1c.
 *
 * Note that P1003.1c specifies that there are no defined comparison
 * or assignment operators for the types pthread_attr_t, pthread_cond_t,
 * pthread_condattr_t, pthread_mutex_t, pthread_mutexattr_t.
 */
typedef struct	pthread			*pthread_t;
typedef struct	pthread_attr		*pthread_attr_t;
typedef struct	pthread_mutex		*pthread_mutex_t;
typedef struct	pthread_mutex_attr	*pthread_mutexattr_t;
typedef struct	pthread_cond		*pthread_cond_t;
typedef struct	pthread_cond_attr	*pthread_condattr_t;
typedef int     			pthread_key_t;
typedef struct	pthread_once		pthread_once_t;
typedef struct	pthread_rwlock		*pthread_rwlock_t;
typedef struct	pthread_rwlockattr	*pthread_rwlockattr_t;

/*
 * Additional type definitions:
 *
 * Note that P1003.1c reserves the prefixes pthread_ and PTHREAD_ for
 * use in header symbols.
 */
typedef void	*pthread_addr_t;

/* TODO : fix this typedef! */
//typedef void	*(*pthread_startroutine_t) (void *);
typedef int32 (*pthread_startroutine_t) (void *);

//typedef thread_func pthread_startroutine_t ;

/*
 * Once definitions.
 */
struct pthread_once {
	int		state;
	pthread_mutex_t	mutex;
};

/*
 * Flags for once initialization.
 */
#define PTHREAD_NEEDS_INIT  0
#define PTHREAD_DONE_INIT   1

/*
 * Static once initialization values. 
 */
#define PTHREAD_ONCE_INIT   { PTHREAD_NEEDS_INIT, NULL }

/*
 * Static initialization values. 
 */
#define PTHREAD_MUTEX_INITIALIZER	NULL
#define PTHREAD_COND_INITIALIZER	NULL
#define PTHREAD_RWLOCK_INITIALIZER	NULL

/*
 * Default attribute arguments (draft 4, deprecated).
 */

extern struct pthread_attr pthread_attr_default;
extern struct pthread_cond_attr pthread_condattr_default;
extern struct pthread_mutex_attr pthread_mutexattr_default;  

enum pthread_mutextype {
	PTHREAD_MUTEX_DEFAULT		= 1,
	PTHREAD_MUTEX_RECURSIVE		= 2,
	PTHREAD_MUTEX_NORMAL		= 3,
	PTHREAD_MUTEX_ERRORCHECK	= 4,
	MUTEX_TYPE_MAX
};

#define MUTEX_TYPE_FAST			PTHREAD_MUTEX_DEFAULT
#define MUTEX_TYPE_COUNTING_FAST	PTHREAD_MUTEX_RECURSIVE

/*
 * Thread function prototype definitions:
 */
int		pthread_attr_destroy (pthread_attr_t *);
int		pthread_attr_getinheritsched (pthread_attr_t *, int *);
int		pthread_attr_getschedparam (pthread_attr_t *,
			struct sched_param *);
int		pthread_attr_getschedpolicy (pthread_attr_t *, int *);
int		pthread_attr_getscope (pthread_attr_t *, int *);
int		pthread_attr_getstacksize (pthread_attr_t *, size_t *);
int		pthread_attr_getstackaddr (pthread_attr_t *, void **);
int		pthread_attr_getdetachstate (pthread_attr_t *, int *);
int		pthread_attr_init (pthread_attr_t *);
int		pthread_attr_setinheritsched (pthread_attr_t *, int);
int		pthread_attr_setschedparam (pthread_attr_t *,
			struct sched_param *);
int		pthread_attr_setschedpolicy (pthread_attr_t *, int);
int		pthread_attr_setscope (pthread_attr_t *, int);
int		pthread_attr_setstacksize (pthread_attr_t *, size_t);
int		pthread_attr_setstackaddr (pthread_attr_t *, void *);
int		pthread_attr_setdetachstate (pthread_attr_t *, int);
void		pthread_cleanup_pop (int execute);
void		pthread_cleanup_push (void (*routine) (void *),
			void *routine_arg);
int		pthread_condattr_destroy (pthread_condattr_t *attr);
int		pthread_condattr_init (pthread_condattr_t *attr);
int		pthread_condattr_getpshared (pthread_condattr_t *attr,
			int *pshared);
int		pthread_condattr_setpshared (pthread_condattr_t *attr,
			int pshared);
int		pthread_cond_broadcast (pthread_cond_t *);
int		pthread_cond_destroy (pthread_cond_t *);
int		pthread_cond_init (pthread_cond_t *,
			const pthread_condattr_t *);
int		pthread_cond_signal (pthread_cond_t *);
int		pthread_cond_timedwait (pthread_cond_t *,
			pthread_mutex_t *, const struct timespec * abstime);
int		pthread_cond_wait (pthread_cond_t *, pthread_mutex_t *);
int		pthread_create (pthread_t *thread, const pthread_attr_t *attr,
			pthread_startroutine_t start_routine, void *arg);
int		pthread_detach (pthread_t thread);
int		pthread_equal (pthread_t, pthread_t);
void		pthread_exit (void *status);
void		*pthread_getspecific (pthread_key_t);
int		pthread_join (pthread_t, void **);
int		pthread_key_create (pthread_key_t *,
			void (*routine) (void *));
int		pthread_key_delete (pthread_key_t);
int		pthread_kill (struct pthread *, int);
int		pthread_mutexattr_destroy (pthread_mutexattr_t *);
int		pthread_mutexattr_getprioceiling (pthread_mutexattr_t *,
			int *prioceiling);
int		pthread_mutexattr_getprotocol (pthread_mutexattr_t *,
			int *protocol);
int		pthread_mutexattr_getpshared (pthread_mutexattr_t *,
			int *pshared);
int		pthread_mutexattr_init (pthread_mutexattr_t *);
int		pthread_mutexattr_setprioceiling (pthread_mutexattr_t *,
			int prioceiling);
int		pthread_mutexattr_setprotocol (pthread_mutexattr_t *,
			int protocol);
int		pthread_mutexattr_setpshared (pthread_mutexattr_t *,
			int pshared);
int		pthread_mutexattr_settype (pthread_mutexattr_t *, int);
int		pthread_mutex_destroy (pthread_mutex_t *);
int		pthread_mutex_getprioceiling (pthread_mutex_t *);
int		pthread_mutex_init (pthread_mutex_t *,
			const pthread_mutexattr_t *);
int		pthread_mutex_lock (pthread_mutex_t *);
int		pthread_mutex_setprioceiling (pthread_mutex_t *);
int		pthread_mutex_trylock (pthread_mutex_t *);
int		pthread_mutex_unlock (pthread_mutex_t *);
int		pthread_once (pthread_once_t *,
			void (*init_routine) (void));

int		pthread_rwlock_init_np (pthread_rwlock_t *,
			const pthread_rwlockattr_t *);
int		pthread_rwlock_destroy_np (pthread_rwlock_t *);
int		pthread_rwlock_rdlock_np (pthread_rwlock_t *);
int		pthread_rwlock_tryrdlock_np (pthread_rwlock_t *);
int		pthread_rwlock_trywrlock_np (pthread_rwlock_t *);
int		pthread_rwlock_unlock_np (pthread_rwlock_t *);
int		pthread_rwlock_wrlock_np (pthread_rwlock_t *);
int		pthread_rwlockattr_init_np (pthread_rwlockattr_t *);
int		pthread_rwlockattr_getpshared_np (const pthread_rwlockattr_t *,
			int *);
int		pthread_rwlockattr_setpshared_np (pthread_rwlockattr_t *,
			int *);
int		pthread_rwlockattr_destroy_np (pthread_rwlockattr_t *);

pthread_t	pthread_self (void);
int		pthread_setcancelstate (int, int *);
int		pthread_setcanceltype (int, int *);
int		pthread_setspecific (pthread_key_t, const void *);
int		pthread_sigmask (int, const sigset_t *, sigset_t *);
int		pthread_testcancel (void);


int		pthread_getprio (pthread_t);
int		pthread_setprio (pthread_t, int);
void		pthread_yield (void);
int		pthread_setschedparam (pthread_t pthread, int policy,
			struct sched_param * param);
int		pthread_getschedparam (pthread_t pthread, int *policy,
			struct sched_param * param);
int		pthread_attr_setfloatstate (pthread_attr_t *, int);
int		pthread_attr_getfloatstate (pthread_attr_t *, int *);
int		pthread_attr_setcleanup (pthread_attr_t *,
			void (*routine) (void *), void *);

#endif
