/* routines to create/ destroy threads etc... */
#include "pthread.h"
#include "pthread_private.h"
#include <malloc.h>
#include <stdio.h>
#include "pthread_debug.h"

/* global default initialisers */
struct pthread_attr pthread_attr_default = {
  PTHREAD_DEFAULT_PRIORITY,
  PTHREAD_CREATE_RUNNING,
  PTHREAD_CREATE_JOINABLE, 
  NULL
};

struct pthread_cond_attr pthread_condattr_default = { COND_TYPE_FAST , 0 };


int		pthread_create (pthread_t *thread, const pthread_attr_t *attr,
		pthread_startroutine_t start_routine, void *arg)
{
	pthread_t		new_thread;
	pthread_attr_t	pattr;
	int rtn = 0;
	int32 startprio;
	thread_info info;

	/* is the reqdy queue initialized???? */
	if (readyq_initialized != 1)
		_init_globals();
		
    /* this is only for getting the priority of parent thread */
//	get_thread_info(0, &info);

	/* we create and manipulate new_thread then copy it into
	 * thread before returning if all succesful */
	if ((new_thread = malloc(sizeof(struct pthread))) == NULL) {
	    /* we coulnd't get any memory for a new pthread ! */
	    pdbgprint ("%s::Couldn't allocate memory for new thread struct\n",__PRETTY_FUNCTION__);
	    return EAGAIN;
	}

	/* now do we need to create attributes ? */
	if (attr == NULL || *attr==NULL) {
		/* yes - use defaults */
		pattr = &pthread_attr_default;
	} else {
		/* no - use supplied */
		pattr = *attr;
	}

	memset (new_thread,0,sizeof (struct pthread));

	if (pattr->flags & PTHREAD_INHERIT_SCHED) {
	    startprio = info.priority;
	} else {
		startprio = B_NORMAL_PRIORITY;
	}

	if ((rtn = spawn_thread( start_routine,"pThread", startprio, arg)) != B_NO_ERROR) {
		/* we got a new thread...now allocate the various bits we need */
		/*new_thread->sigmask = ??? inherited from calling thread...*/
    	memcpy(&new_thread->attr, pattr, sizeof(struct pthread_attr));
		new_thread->tid = rtn;
		if (new_thread->attr.flags & PTHREAD_INHERIT_SCHED) {
			/* we need to use parent thread scheduling... */
			new_thread->base_priority = startprio;
		} else {
			/* we just use the attribute priority */
			new_thread->base_priority = new_thread->attr.priority;
		}
		new_thread->active_priority = new_thread->base_priority;

		/* initialize the queue of people waiting to join us */
		_thread_queue_init(&(new_thread->join_queue));

		_lock_thread_list();	
		new_thread->nxt = _thread_link_list;
		_thread_link_list = new_thread;
		_unlock_thread_list();
		
		if (pattr->suspend == PTHREAD_CREATE_SUSPENDED) {
			new_thread->state = PS_SUSPENDED;
			PTHREAD_WAITQ_INSERT(new_thread);
			/* we're suspended so don't start running yet ! */
		} else {
			new_thread->state = PS_RUNNING;
			PTHREAD_PRIOQ_INSERT_TAIL(new_thread);
			/* well alright we're running - so kick it off ! */
			resume_thread(rtn);
		}		
    } else {
		if ((rtn==B_NO_MORE_THREADS) || (rtn==B_NO_MEMORY))
		    pdbgprint ("%s::No more threads or memory\n",__PRETTY_FUNCTION__);
			return (EAGAIN);
	}

	(*thread) = new_thread;
	return (0);
}

void		pthread_exit (void *status)
{
	/*
		terminates the calling thread, return value is in status.
		This is supposed to pop and execute any cancelation cleanup handlers,
		the call any TSD destructor funcs.
	*/
	status_t r=B_OK,*ret=NULL;
	/* make sure that we don't derefence a NULL pointer. */
	ret=(status==NULL)?&r:status;
	
	exit_thread (*ret);
}

void _init_globals() {
	/* this needs to be called if we haven't yet initialized the queues... */
	readyq_initialized = 1;
	_pq_init(&_readyq, PTHREAD_MIN_PRIORITY, PTHREAD_MAX_PRIORITY);
}


int	pthread_detach (pthread_t thread)
{
	pthread_t thread1;
	if (thread == NULL) {
	    pdbgprint ("%s::passed NULL thread struct\n",__PRETTY_FUNCTION__);
		return EINVAL;
	}
	/* we need to kick everyone off our join_queue */
	while ((thread1 = _thread_queue_deq(&thread->join_queue)) != NULL ) {
		PTHREAD_NEW_STATE(thread1, PS_RUNNING);
		resume_thread(thread1->tid);
	}
	return kill_thread (thread->tid);
}

int pthread_join(pthread_t thrd, void ** retval) {
	struct pthread *me = NULL;
	/* basically we can just wait until it's dead! */
	if (thrd == NULL) {
		/* if there is no pthread then just return */
	    pdbgprint ("%s::passed NULL thread struct\n",__PRETTY_FUNCTION__);
		return EINVAL;
	}
	if (thrd->tid == find_thread(NULL)) {
		/* this prevents a deadlock of trying to wait on ourselves!! */
	    pdbgprint ("%s::attempted join with the same thread\n",__PRETTY_FUNCTION__);
		return EINVAL;
	}
	/* we need to add ourselves to the threads queue */
	/* we probably should do something in case of an error :-( */
	if ((me = _find_self()) != NULL)
		/* OK so we're not ourselves a pthread...but this isn't an issue! */
		_thread_queue_enq(&(thrd->join_queue), me);
	
	errno=wait_for_thread(thrd->tid, (void*)&retval);
	if (errno != B_NO_ERROR) {
	    pdbgprint ("%s::wait_for_thread returned: 0x%x:%s\n",__PRETTY_FUNCTION__,errno,strerror (errno));
		return errno;
	}

	return 0;
	
}

int		pthread_once (pthread_once_t *once,void (*init_routine) (void))
{
	if (once->state==PTHREAD_NEEDS_INIT) {
		pthread_mutex_lock (&once->mutex);
		if (once->state==PTHREAD_NEEDS_INIT) {
			init_routine ();
			once->state=PTHREAD_DONE_INIT;
		};
		pthread_mutex_unlock (&once->mutex);
	};
	return (EOK);
}

void		pthread_yield (void)
{
	/* Hmm, sleep and let the scheduler yeild ya.  this sleeps for 2 ms.*/
	snooze (2);
}

struct pthread * _find_self() {
	struct pthread * thread1;
	thread_id me = find_thread(NULL);

	_lock_thread_list();	
	thread1 = _thread_link_list;
	while (thread1 != NULL && thread1->tid != me) {
		thread1 = thread1->nxt;
	}
	_unlock_thread_list();

	return (thread1);
}

int _find_thread(struct pthread * thread) {
	struct pthread * thread1;
	
	_lock_thread_list();
	thread1 = _thread_link_list;
	while (thread1 != NULL && thread1 != thread)
		thread1 = thread1->nxt;
	_unlock_thread_list();
	return ((thread1 != NULL) ? 0 :	ESRCH);
}


	
